/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; c-file-offsets: ((substatement-open . 0)) -*- */
/**
 *
 * @file	Location.hpp
 * @author	Ingo Luetkebohle (iluetkeb@TechFak.Uni-Bielefeld.de)
 * @date	started at 2004/06/18
 * @version	$Id: Location.hpp 4273 2016-09-06 12:44:18Z rhaschke $
 *
 **/

#ifndef __XMLTIO_LOCATION_HPP
#define __XMLTIO_LOCATION_HPP

#include "str.hpp"
#include "exception.hpp"
#include "Namespaces.hpp"
#include "Document.hpp"
#include "ID.hpp"
#include "xmltio_config.hpp"

#include <string>
#include <sstream>
#include <xercesc/dom/DOMNode.hpp>
#include <xqilla/items/Node.hpp>
class DynamicContext;

#include <stdlib.h>		// for strtol

namespace xmltio {
// forward declarations
class XMLTIO_API XPath;
class XMLTIO_API Location;

/** \brief return the name of the given element type,
 * or "unknown".
 */
const char* typestring(int type);

/** \brief A wrapper class for the content of a Location.
 *
 * The assignment Location = Location can have two meanings:
 * a) assignment on the class level (flat copy in this case)
 * b) assignment on the content level, i.e. replace the content
 *    of the first location by the content of the second.
 * To resolve this ambiguity, the wrapper LocationContent is
 * introduced. To achieve b) simply write:
 * Location = LocationContent (Location)
 */

class XMLTIO_API LocationContent {
private:
	const Location& _loc;

	/** Disable assignment operator */
	LocationContent& operator=(const LocationContent&);

public:
	LocationContent(const Location& loc) : _loc(loc) {}

	friend class Location;
	friend std::ostream& operator<<(std::ostream&, const LocationContent&lc);
};

/** \brief A writable location in a Document.
 *
 * Typically specified by an XPath expression, this class
 * allows convenient access to a particular node in a document.
 *
 * Assumes that the <em>structure</em> of the document already
 * exists and allows modification.  Attributes can be added,
 * other structure building is beyond the scope of this class.
 *
 **/
class XMLTIO_API Location {
protected:
	typedef XERCES_CPP_NAMESPACE::DOMNode Node;
	TIODocument doc_;
	Node *node_;

	static Node* get_xpath(const Location&, const XPath&);
	/** \brief return named attribute, or NULL if not present.
	 */
	Node* attr(const char *name) const;

	Node* import(const Location& l);
public:
	/** \brief internal constructor */
	Location(const TIODocument& doc, Node* n=NULL);

	// copy ctor
	Location(const Location& l) :
		doc_(l.doc_), node_(l.node_) {
	}

	/** \brief create location for the document given.
	 *
	 * @param xml an XML document to parse
	 * @param xpath path into the document
	 */
	Location(const std::string& xml, const XPath& xpath);

	/** \brief Create location from xml document, using XPath("/")
	 *
	 */
	explicit Location(const std::string& xml);

	/** \brief Create location from XPath expression.
	 *
	 * If more than one node matches, a random one will be
	 * used. (usually the the first, but that depends on
	 * the XPath evaluators opinion of ANY_UNORDERED_NODE).
	 *
	 * @param ctx context node for XPath
	 * @param xpath XPath expression
	 * @throws MalformedXPathException if expression is invalid
	 * @throws NotFoundException if no matching node exists. */
	Location(const Location& ctx, const XPath& xpath);

	/** \brief Create location from TIODocument and XPath expression.
	 *
	 * If more than one node matches, a random one will be
	 * used. (usually the the first, but that depends on
	 * the XPath evaluators opinion of ANY_UNORDERED_NODE).
	 *
	 * @param d TIODocument as context for XPath
	 * @param xpath XPath expression
	 * @throws MalformedXPathException if expression is invalid
	 * @throws NotFoundException if no matching node exists. */
	Location(const TIODocument& d, const XPath& xpath);

	/** \brief return named attribute, creating it if necessary.
	 *
	 * @return Location instance for the attribute node.
	 * @throws UnsupportedOperationException
	 *   if the node type can't accept attributes.
	 */

	Location operator[](const char* name);
	/** \brief return named attribute, if it exists.
	 *
	 * @return Location instance for
	 * @throws NotFoundException
	 *   if the attribute does not exist
	 * @throws UnsupportedOperationException
	 *   if this node type can't accept attributes
	 */
	Location operator[](const char* name) const;

	Location operator[](const XPath& xpath) const;

	/** \brief return name of the node */
	std::string getName() const {
		const XMLCh* name = node_->getLocalName();
		if(!name) {
			return NX(node_->getNodeName());
		}
		return NX(name);
	}

	/** return the xml:id attribute of this element, if any.
	 * For a unique id of the whole document, see
	 * TIODocument.getID().
	 *
	 * @throw IDNotFoundException if this node does not or cannot have an xml:id attribute.
	 * @see TIODocument.getID()
	 */
	ID getID() const {
		try {
			return (*this)["xml:id"].getTextValue();
		} catch(const XMLTIO_Exception& e) {
			throw IDNotFoundException(e);
		}
	}
	/**
	 * \brief Assign an ID to this element.
	 */
	void setID(const ID& id);

	/** \brief Shallow copy of Location.
	 *
	 * Afterwards this and l refer to the same location
	 * of the very same document, i.e. changes apply to
	 * both of them equally.
	 */
	Location& operator=(const Location& l)
		{
			node_ = l.node_;
			doc_  = l.doc_;
			return *this;
		}

	/** \brief Content copy/replacement of Location.
	 *
	 * Replace content by given LocationContent.
	 */
	Location& operator=(const LocationContent& lc)
		{
			return this->replace (lc._loc);
		}

	/** \brief set text for node.
	 *
	 * If this node is an attribute, text or character
	 * data node, it will be modified directly.
	 * Otherwise, its TextContent will be set.
	 **/
	Location& operator=(const std::string& val);

	/** \brief set value by converting argument to string.
	 *
	 * A converter function "convert" for the right type
	 * must exist.  Users can supply their own
	 * specializations for new types.
	 *
	 * @param value new node value
	 * @see convert
	 */
	template<typename T>
		Location& operator=(const T& value) {
		try {
			convert(value, *this);
			return *this;
		} catch(const XERCES_CPP_NAMESPACE::DOMException& e) {
			throw XMLTIO_Exception(NX(e.msg));
		}
	}

	/** \brief retrieve node value as text.
	 *
	 * For attribute, text or character data node,
	 * the node value is returned.  For others, the
	 * TextContent is returned.
	 *
	 */
	std::string getTextValue() const;

	/** \brief return document this location is in.
	 */
	TIODocument& getDocument() {
		return doc_;
	}
	const TIODocument& getDocument() const {
		return doc_;
	}

	std::string getDocumentText(bool pretty=false) const {
		return dom2string(doc_, pretty);
	}
	std::string getText(bool pretty=false) const {
		return dom2string(node_, pretty);
	}

	bool operator!=(const Location& o) const
	{
		return node_ != o.node_;
	}
	bool operator==(const Location& o) const
	{
		return node_ == o.node_;
	}

	bool isEqual(const Location& o) const
	{
		return node_->isEqualNode(o.node_);
	}

	/** \brief return parent location */
	Location getParent()
	{
		return Location(doc_, node_->getParentNode());
	}

	/// append a new (empty) element and return as Location
	Location createElement(const char* pcTag);
	/// append a new (empty) element and return as Location
	Location createElement(const std::string& sTag) {
		return createElement(sTag.c_str());
	}

	/** \brief Add a copy of the Location
	 *
	 * Returns the new Location.
	 */
	Location add(const Location& l);

	/** \brief Add a copy of the Location before reference
	 *
	 * Does the same as Add but inserts the Location before a
	 * reference child Location. If the Location is an Attribute
	 * the behaviour will be exactly the same as with add.
	 */
	Location addBefore(const Location& l, const Location& lref);

	/** \brief remove given child Location.
	 *
	 * removes child Location from this tree and
	 * unsets the Locations' document.
	 * @throws XMLTIO_Exception if l is not a child of this node.
	 */
	void remove(Location& l);

	/** \brief replace content by content of given location.
	 *
	 * The new node will be copied before being inserted into the
	 * tree.  If 'old' is not a child, an exception will
	 * be raised.
	 *
	 * @param n new content
	 * @return the changed location
	 */
	Location& replace(const Location& n);

	static ::Node::Ptr createXQNode(XERCES_CPP_NAMESPACE::DOMNode* node, DynamicContext* context);

	/** \brief create a XQilla Node* from this location
	 *
	 *  The XQilla::XercesNodeImpl* is a wrapper around the
	 *  existing DOMNode* which is suitable for XQilla
	 */
	::Node::Ptr createXQNode(DynamicContext& context) {
		return createXQNode(node_, &context);
	}

	friend class XPath;
	friend class TIODocument;

	friend std::ostream& operator<<(std::ostream&, const Location&l);
};
typedef boost::shared_ptr<Location> LocationPtr;
/** \brief return a location or an empty pointer if it does
    not exist. */
XMLTIO_API LocationPtr find(const Location&, const XPath&);

// std output operator
inline std::ostream& operator<<(std::ostream& out,
                                const Location &l)
{
	return out << dom2string(l.node_);
}
}

#endif //  __XMLTIO_LOCATION_HPP
