/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; c-file-offsets: ((substatement-open . 0)) -*- */

#include "Namespaces.hpp"
#include "str.hpp"
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMAttr.hpp>
#include <xercesc/util/XMLString.hpp>

#include <sstream>

using namespace std;
using namespace XERCES_CPP_NAMESPACE;

namespace xmltio {
Namespaces::Namespaces(DOMElement* de)
	: de_(de)
{
	if(!de)
		throw XMLTIO_Exception("document element NULL");
}

/** \brief add namespace mapping from prefix to uri.
 *
 * Works by adding attributes to the contained element.  If a
 * mapping for that prefix already exists, its value is
 * checked against the given URI.
 *
 * @param prefix the prefix to be mapped
 * @param uri target of the mapping
 * @throws InvalidCharacterException when prefix or uri contain illegal characters
 * @throws XMLTIO_Exception when prefix is already present with a different URI
 */
void Namespaces::put(const string& prefix, const string& uri)
{
	try {
		const X qname("xmlns:" + prefix);
		const X xuri(uri);

		// attribute already present?  make sure it
		// has the same uri value
		// otherwise, complain
		const XMLCh* prev(de_->getAttribute(qname));
		if(NULL != prev && '\0' != prev[0]) {
			if(XMLString::compareString(prev, xuri) == 0) {
				return;
			}

			ostringstream msg;
			msg << "prefix '" << prefix
			    << "' already present with uri '"
			    << NX(prev)
			    << "' (you requested " << uri
			    << ')';
			throw XMLTIO_Exception(msg.str());
		}

		// add new
		de_->setAttributeNS(X("http://www.w3.org/2000/xmlns/"),
		                    qname, xuri);
	} catch(const DOMException& ex) {
		ostringstream msg;
		msg << "couldn't add namespace declaration("
		    << prefix << '=' << uri << ')';
		throw InvalidCharacterException(msg.str(), ex);
	}
}

void Namespaces::remove(const std::string& prefix)
{
	try {
		de_->removeAttribute(X("xmlns:" + prefix));
	} catch(const DOMException& ex) {
		throw NotFoundException(ex);
	}
}

void Namespaces::merge(const Namespaces& o)
{
	DOMNamedNodeMap* attrs(o.de_->getAttributes());
	for(unsigned int i = 0; i < attrs->getLength(); ++i) {
		DOMNode* node(attrs->item(i));
		if(!node)
			continue;
		// ignore attrs not starting with "xmlns:"
		string name((const char*)NX(node->getNodeName()));
		if(name.find("xmlns:") == string::npos)
			continue;

		put(name.substr(name.find(":") +1),
		    NX(node->getNodeValue()));
	}
}
}
