#include "XPath.hpp"
#include "Location.hpp"
#include "log.hpp"
#include "exception.hpp"
#include "xinit.hpp"
#include "str.hpp"

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/sax/InputSource.hpp>
#include <xercesc/framework/MemBufFormatTarget.hpp>

#include <xqilla/exceptions/XQException.hpp>
#include <xqilla/utils/UTF8Str.hpp>
#include <xqilla/xerces/XercesConfiguration.hpp>
#include <xqilla/dom-api/impl/XQillaNSResolverImpl.hpp>
#include <xqilla/context/ContextHelpers.hpp>
#include <xqilla/framework/ProxyMemoryManager.hpp>
#undef X

#include <iostream>

using namespace std;
using namespace XERCES_CPP_NAMESPACE;

GETLOGGER("xmltio.xpath");

namespace xmltio {

void XPathIterator::next() {
	itm_ = res_->next(ctx_.get());
}

XPathIterator::Node* XPathIterator::item() const {
	if (!itm_) return NULL;
	return (XPathIterator::Node*) itm_->getInterface(XercesConfiguration::gXerces);
}

XPathIterator::XPathIterator() : res_(0) {}

/// copy constructor destroys old result instance,
/// such that only the current item will be returned
/// TODO: Can't we use both xpath iterators indepedently?
XPathIterator::XPathIterator(const XPathIterator& o)
	: doc_(o.doc_), ctx_(o.ctx_), res_(o.res_), itm_(o.itm_) {}

XPathIterator::XPathIterator(const TIODocument& doc, const Result& r,
                             const DynamicContextPtr& c)
	: doc_(doc), ctx_(c), res_(r)
{
	next();
}

XPathIterator::~XPathIterator() {}

XPathIterator& XPathIterator::operator=(const XPathIterator& o)
{
	doc_   = o.doc_;
	res_   = o.res_;
	itm_   = o.itm_;
	// rhaschke: it is important to assign context last -- otherwise segfaults are possible
	ctx_   = o.ctx_;
	return *this;
}

static XQillaNSResolverImpl* gpNSResolver = 0;
static const XQillaNSResolver* getNSResolver () {
	if (!gpNSResolver) {
		static const char* apcNameSpaceList[][2] =
			{{"dbxml", "http://www.sleepycat.com/2002/dbxml"},
			 {"memory", "http://purl.org/ActiveMemory"},
			 {NULL, NULL}};
		ProxyMemoryManager* memMgr = new ProxyMemoryManager(
			XMLPlatformUtils::fgMemoryManager);
		gpNSResolver = new XQillaNSResolverImpl(memMgr);
		for (int i=0; apcNameSpaceList[i][0]; ++i) {
			gpNSResolver->addNamespaceBinding (X(apcNameSpaceList[i][0]),
			                                   X(apcNameSpaceList[i][1]));
		}
	}
	return gpNSResolver;
}

XPath::XPath(const char* xpath) : xpath_(xpath)
{
	if(!xpath || !*xpath) {
		throw MalformedXPathException("empty xpath");
	}
	pthread_mutex_init(&contextMutex_, NULL);
	compile();
}
XPath::XPath(const std::string& xpath) : xpath_(xpath)
{
	if("" == xpath) {
		throw MalformedXPathException("empty xpath");
	}
	pthread_mutex_init(&contextMutex_, NULL);
	compile();
}
XPath::XPath(const XPath& o)
{
	pthread_mutex_init(&contextMutex_, NULL);
	*this = o;
}
XPath::~XPath() {
	pthread_mutex_destroy(&contextMutex_);
}

bool XPath::isMatch(const Location& l) const
{
	if(!*this) {
		return false;
	}

	XPathIterator i(evaluate(l));
	return !i == false;
}

void XPath::compile () {
	static ::XercesConfiguration config;
#if XQILLA_VERSION >= 0203
	// bool conditions fail with XQilla 2.3 and XPATH2 language
	// example: "/*[not(@arg)]"
	const int language = XQilla::XQUERY;
#else
	const int language = XQilla::XPATH2;
#endif
	if (!context_) { // reuse an existing context
		context_ = DynamicContextPtr(
			getXQ().createContext((XQilla::Language)language, &config));
#if XQILLA_VERSION != 0203 // impedes usage of functions in XQilla 2.3
		context_->setNSResolver(getNSResolver());
#endif
	}

	try {
		query_ = XQQueryPtr(getXQ().parse(X(xpath_), context_.get(), 0,
		                                  XQilla::NO_ADOPT_CONTEXT));
	} catch (const XQException &e) {
		// reset a previous query in any case
		if (query_) query_.reset();

		// filter errors regarding unknown namespace prefixes:
		// namespace will be defined in document only:
		// requires dynamic parsing of XPath if doc is available
		const string error (UTF8(e.getError()));
		DEBUG("static compilation failed for '" << xpath_ << "' : " << error);
		if (error.find("[err:XPST0081]") == string::npos)
			throw MalformedXPathException (error);
	}
}

XPathIterator XPath::evaluate(const TIODocument& doc,
                              DOMNode* node) const
{
	if(!node) {
		throw XMLTIO_Exception("context node NULL");
	}

	// set up XPath engine
	try {
		pthread_mutex_lock(&contextMutex_);

		bool bDynamicQuery = !query_;
		if (bDynamicQuery) {
			// compile query with current document as namesspace scope
			XQillaNSResolverImpl nodeNSScope(context_->getMemoryManager(),
			                                 node, context_->getNSResolver());
			AutoNsScopeReset jan(context_.get(),&nodeNSScope);

			// TOOD: using a NS resolver impedes the usage of functions in XQilla 2.3
			query_ = XQQueryPtr(getXQ().parse(X(xpath_), context_.get(), 0,
			                                  XQilla::NO_ADOPT_CONTEXT));
		}
		/* We use the same context for each evaluation, because context
		   creation is *the* expensive operation (nearly half of the time) */
		context_->setContextItem(Location::createXQNode(node, context_.get()));
		context_->setContextPosition(1);
		context_->setContextSize(1);

		XPathIterator result (doc, query_->execute(context_.get()), context_);
		context_->setContextItem(0);

		pthread_mutex_unlock(&contextMutex_);

		/* Reevaluate query in next document context to ensure
		   that we use those namespace bindings */
		if (__builtin_expect(bDynamicQuery, false)) {query_.reset();}
		return result;
	} catch (const XQException &e) {
		const string error (UTF8(e.getError()));
		// filter errors regarding unknown namespace prefix
		if (error.find("[err:XPST0081]") != string::npos)
			return XPathIterator(doc, Result(0), DynamicContextPtr());
		else throw XMLTIO_Exception(error);
	}
}

// FIXME: NAMESPACE SUPPORT
XPathIterator XPath::evaluate(const Location& l) const
{
	Location& lnc = const_cast<Location&>(l);
	return evaluate(lnc.getDocument(), lnc.node_);
}

std::ostream& operator<<(std::ostream& out, const XPath& xp) {
	out << xp.xpath_;
	return out;
}
std::istream& operator>>(std::istream& in, XPath& xp) {
	string xpath;
	in >> xpath;
	xp = xpath;
	return in;
}

void convert(const XPath& o, Location& l)
{
	l = o.xpath_;
}
}
