/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; c-file-offsets: ((substatement-open . 0)) -*- */

#include "XercesDocumentBody.hpp"
#include "Location.hpp"
#include "log.hpp"

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMException.hpp>

#include <memory>

using namespace XERCES_CPP_NAMESPACE;
using namespace std;

GETLOGGER("xmltio.XercesDocumentBody");

namespace {
using namespace xmltio;

ostream& operator<<(ostream& out, const XMLCh* msg)
{
	out << NX(msg);
	return out;
}

class StrictErrorHandler : public HandlerBase {
	virtual void error(const SAXParseException& exc)
	{
		DEBUG("xml parser error: " << exc.getMessage());
		throw exc;
	};
	virtual void warning(const SAXParseException& exc)
	{
		DEBUG("xml warning: " << exc.getMessage());
	}
	virtual void fatalError(const SAXParseException& exc)
	{
		DEBUG("fatal parser error: " << exc.getMessage());
		throw exc;
	}
};

DOMDocument* do_parse(const std::string& xml)
{

	if(0 == xml.size() || "" == xml) {
		throw ParsingException("empty document");
	}
	try {
		auto_ptr<StrictErrorHandler>
			errh(new StrictErrorHandler());
		auto_ptr<XercesDOMParser>
			parser(new XercesDOMParser());

		parser->setDoNamespaces(true);
		parser->setErrorHandler(errh.get());

		MemBufInputSource input((unsigned char*)xml.c_str(),
		                        xml.length(),
		                        "", false);
		parser->parse(input);

		DOMDocument* doc = parser->adoptDocument();
		if(!doc) {
			throw ParsingException("empty document");
		}
		parser->reset();
		return doc;
	} catch(const DOMException& e) {
		throw ParsingException(NX(e.msg), xml);
	} catch(const XMLException& e) {
		throw ParsingException(NX(e.getMessage()), xml);
	} catch(const SAXParseException& ex) {
		throw ParsingException(NX(ex.getMessage()), xml);
	}
}

DOMDocument* docFromLocation(DOMDocument* oldDoc, DOMNode* n)
{
	try {
		DOMDocument* newDoc = (DOMDocument*)oldDoc->cloneNode(false);
		DOMNode* node = newDoc->importNode(n, true);
		newDoc->insertBefore(node, NULL);
		return newDoc;
	} catch(const DOMException& ex) {
		throw UnsupportedOperationException(ex);
	}
}

}


namespace xmltio {

XercesDocumentBody::XercesDocumentBody(const std::string& xml)
	: doc_(do_parse(xml)), nsps_(doc_->getDocumentElement())
{
}
XercesDocumentBody::XercesDocumentBody(DOMDocument* doc)
	: doc_(doc), nsps_(doc_->getDocumentElement())
{
}
XercesDocumentBody::XercesDocumentBody(const XercesDocumentBody& o)
	: doc_((DOMDocument*)o.doc_->cloneNode(true)),
	  nsps_(doc_->getDocumentElement())
{
}
XercesDocumentBody::XercesDocumentBody(const XercesDocumentBody& o,
                                       DOMNode *subpart)
	: doc_(docFromLocation(const_cast<DOMDocument*>(o.doc_),
	                       subpart)),
	  nsps_(doc_->getDocumentElement())
{
}
XercesDocumentBody::~XercesDocumentBody()
{
	doc_->release();
}

ostream& operator<<(ostream& out, const XercesDocumentBody& doc)
{
	out << dom2string(doc.doc_);
	return out;
}


}
