/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; c-file-offsets: ((substatement-open . 0)) -*- */

#include <vector>
#include <iostream>
#include <limits>

#include "../src/xmltio.hpp"
#include "../src/log.hpp"
#include "../src/XercesDocumentBody.hpp"

#define BOOST_TEST_MODULE tio_test
#include <boost/test/unit_test.hpp>
#include <boost/test/unit_test_log.hpp>

#include <values.h>

using namespace std;
using namespace xmltio;
using namespace XERCES_CPP_NAMESPACE;
using namespace boost;
using namespace boost::unit_test_framework;

const char* L4CONF = "log4cpp.props";

GETLOGGER("tio_test");

struct TestStruct {
	int x;
	bool operator==(const TestStruct& o) const {
		return x == o.x;
	}
};
ostream& operator<<(ostream& out, const TestStruct& ts)
{
	out << "TestStruct[x=" << ts.x << "]";
	return out;
}

namespace xmltio {
TEMPLATE_TRAITS(TestStruct, "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><test><struct x=\"\"/></test>", "/test/struct");

void convert(const TestStruct& s, Location& l)
{
	DEBUG("convert<TestStruct>(s=" << s << ")");
	l["x"] = s.x;
}
template<>
TestStruct extract<TestStruct>::get() const
{
	DEBUG("extract<TestStruct>[loc_=" << loc_ << "]");
	TestStruct ts;
	ts.x = extract<int>(loc_["x"]);
	return ts;
}
}

const string xmldata = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><xcf><arguments arg=\"v\"><flag>1</flag><flag>2</flag><foo>bar</foo><meier>teich</meier></arguments></xcf>";
const string rdfsample = "<?xml version=\"1.0\"?><rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"><rdf:description rdf:about=\"http://www.vampire-project.de/\"/></rdf:RDF>";
const string xml_attr =  "<?xml version=\"1.0\"?><item attr=\"\"/>";
const string itemdata = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><xcf><item attr=\"\"/></xcf>";
const string id_test = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><item xmlns:dbxml=\"http://www.sleepycat.com/2002/dbxml\" dbxml:id=\"mem_512\"><su xml:id=\"subitem\"/></item>";

void xmltio_translator(const XMLTIO_Exception& e)
{
	BOOST_ERROR(std::string("caught XMLTIO_Exception ") +
	            e.what());
}

BOOST_AUTO_TEST_CASE(test_id)
{
	Location root(id_test, "/");
	Location l(root, "item/su");
	BOOST_CHECK_EQUAL(l.getDocument().getID(), "mem_512");
	BOOST_CHECK_EQUAL(l.getID(), "subitem");
	BOOST_CHECK_THROW(root.getID(), IDNotFoundException);

	l.setID("laber");
	BOOST_CHECK_EQUAL(l.getID(), "laber");
}

BOOST_AUTO_TEST_CASE(test_float)
{
	Location root(xml_attr, "/");
	Location l(root, "item/@attr");

	l=0.5e-20;
	float result = extract<float>(l);
	bool flag = (0.5e-20 - result) <= std::numeric_limits<float>::epsilon();
	BOOST_CHECK(flag);
}

BOOST_AUTO_TEST_CASE(test_str)
{
	// not really a test, but does the job
	cerr << dom2string(parse(xmldata)) << endl;
	BOOST_CHECK_EQUAL(xmldata, dom2string(parse(xmldata)));
	BOOST_CHECK_THROW(parse("this is not xml"), ParsingException);

	// empty document
	BOOST_CHECK_THROW(parse(""), ParsingException);
}

const string xml_nested_text = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><foo>test<bar>baz</bar></foo>";

BOOST_AUTO_TEST_CASE(test_text)
{
	Location l(xml_nested_text, "/foo");
	BOOST_CHECK_EQUAL("test", l.getTextValue());
}


BOOST_AUTO_TEST_CASE(test_location)
{
	// regular use
	Location doc(xmldata, XPath("/xcf"));
	const Location flag(xmldata, XPath("/xcf/arguments/meier"));
	Location f2 = doc.add(flag);
	BOOST_CHECK(find(doc, XPath("/xcf/meier")) != NULL);

	doc.remove(f2);
	BOOST_CHECK(find(doc, XPath("/xcf/meier")) == NULL);
	BOOST_CHECK(find(doc, XPath("/xcf/arguments/meier")) != NULL);

	BOOST_CHECK_EQUAL(doc[XPath("arguments")].getName(),"arguments");

	// typical errors

	// non-existing XPath
	BOOST_CHECK_THROW(doc[XPath("/outofhere")], NotFoundException);
	// non-existing attribute in const location
	BOOST_CHECK_THROW(flag["uhoh"], NotFoundException);
	// bad attribute names
	BOOST_CHECK_THROW(doc[""], InvalidCharacterException);
	BOOST_CHECK_THROW(flag[""], NotFoundException);
	// trying to use an XPath as an attribute
	BOOST_CHECK_THROW(doc["/path"], InvalidCharacterException);
	// assigning to root element
	Location root(xmldata, "/");
	BOOST_CHECK_THROW(root = string("test"),
	                  UnsupportedOperationException);
	// extracting from root element
	int i;
	BOOST_CHECK_THROW(i = extract<int>(root),
	                  UnsupportedOperationException);

	string s;
	BOOST_CHECK_THROW(s = extract<string>(root),
	                  UnsupportedOperationException);
	cout << s << endl;
}

BOOST_AUTO_TEST_CASE(test_XPathIterator)
{
	Location args(xmldata, "/");
	XPathIterator i,j;
	i = XPath("//flag/text()").evaluate(args);
	j = i; // i is transferred to j
	BOOST_CHECK(i); BOOST_CHECK_EQUAL((*i).getTextValue(), "1");
	BOOST_CHECK(j); BOOST_CHECK_EQUAL((*j).getTextValue(), "1");

	++j; BOOST_CHECK(j); BOOST_CHECK_EQUAL((*j).getTextValue(), "2");
	++i; BOOST_CHECK(!i);

	++j; BOOST_CHECK(!j);
}

BOOST_AUTO_TEST_CASE(test_xpath)
{
	BOOST_CHECK_THROW(XPath("[]]]"), MalformedXPathException);

	Location args(xmldata, XPath("/xcf/arguments"));
	BOOST_CHECK_EQUAL(args.getName(), "arguments");

	// check that we find one node with text content '2'
	XPathIterator i = XPath("*[text()=\"2\"]").evaluate(args);
	BOOST_CHECK(i);
	++i;
	BOOST_CHECK(!i);
	i = XPath("*[text()=\"2\"]").evaluate(args);
	XPathIterator foo(i);
	// there should be no node with text content '3'
	i = XPath("*[text()=\"3\"]").evaluate(args);
	XPathIterator bar(i);
	BOOST_CHECK(!i);

	// there should be two 'flag' nodes
	i = XPath("flag").evaluate(args);
	BOOST_CHECK(i);
	++i;
	BOOST_CHECK(i);
	XPathIterator j = ++i;
	BOOST_CHECK(!j);
	BOOST_CHECK(!i);

	i = XPath("@arg").evaluate(args);
	BOOST_CHECK(i);
	string value = extract<string>(*i);
	BOOST_CHECK_EQUAL(value, "v");
	Location arg(args, XPath("@arg"));
	value = extract<string>(arg);
	BOOST_CHECK_EQUAL(value, "v");

}

BOOST_AUTO_TEST_CASE(test_match)
{
	Location l(xmldata, "/");
	XPath nosuchtag("/no/such/tag/here");
	XPath badtext("//foo[text()=\"meier\"]");
	BOOST_CHECK(!nosuchtag.isMatch(l));
	BOOST_CHECK(!badtext.isMatch(l));
}

BOOST_AUTO_TEST_CASE(test_namespace)
{
	Location rdf(rdfsample, "/rdf:RDF");
	Location l = rdf[XPath("rdf:description")];
	BOOST_CHECK_EQUAL(l.getName(), "description");
}

BOOST_AUTO_TEST_CASE(test_find)
{
	XPath xpath("none");
	Location doc(xmldata, XPath("/"));

	DEBUG("find");
	BOOST_CHECK(find(doc, xpath).get() == NULL);
	DEBUG("explicit");
	BOOST_CHECK_THROW(Location(doc, xpath),
	                  NotFoundException);
}

BOOST_AUTO_TEST_CASE(test_set_content)
{
	Location l(xmldata, XPath("/xcf/arguments"));
	const string input("some text");
	l = input;
	string content = extract<string>(l[XPath("text()")]);
	BOOST_CHECK_EQUAL(input, content);
}

template<typename T>
void _test_convert(const T& value, const string& compare)
{
	Location l(xml_attr, "/item/@attr");
	l = value;
	string result = extract<string>(l);
	BOOST_CHECK_EQUAL(compare, result);
}

BOOST_AUTO_TEST_CASE(test_convert)
{
	// constants here are from limits.h
	// we don't use them directly because the full-range
	// is not important, its just important that at least
	// 32 bits are supported

	int i = 2147483647;
	_test_convert(i, "2147483647");

	i = -2147483646;
	_test_convert(i, "-2147483646");

	double d = 1.5;
	_test_convert(d, "1.5");

	d = -d;
	_test_convert(d, "-1.5");
}

BOOST_AUTO_TEST_CASE(test_serialize)
{
	TestStruct s1;
	const int TEST_VAL = 10;
	s1.x = TEST_VAL;
	string xml = to_xml(s1);

	DEBUG("TestStruct to_xml=" << xml);

	Location l(find_anchor<TestStruct>(xml));
	BOOST_CHECK_EQUAL(TEST_VAL, (int)extract<int>(l["x"]));
	TestStruct s2(extract<TestStruct>::from_xml(xml));
	BOOST_CHECK_EQUAL(s1, s2);

	// provoke a conversion error
	l["x"] = string("number 10");
	int i;
	BOOST_CHECK_THROW(i = extract<int>(l["x"]),
	                  ConversionFailedException);

	// empty field is an error
	l["x"] = string("");
	BOOST_CHECK_THROW(i = extract<int>(l["x"]),
	                  ConversionFailedException);
}

BOOST_AUTO_TEST_CASE(test_xml_as_string)
{
	string xml = to_xml(xmldata);
	DEBUG("xml content from cdata: " << xml);
	BOOST_CHECK_EQUAL(xmldata, extract<string>::from_xml(xml));
}


const string bare = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><bare/>";
const string bare_memory_ns = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><bare xmlns:memory=\"http://purl.org/ActiveMemory\" memory:about=\"about something\"/>";
const string bare_memory_ns_no_about = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><bare xmlns:memory=\"http://purl.org/ActiveMemory\"/>";
const string bare_plus_foo = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><bare><foo/></bare>";
const string just_foo = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><foo/>";
const string bare_plus_foo_ns = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><bare xmlns:xcf=\"http://purl.org/XCF\"><xcf:foo/></bare>";
const string rdf_bare = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><bare xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"/>";
const string rdf_desc = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><foo xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"><rdf:description></rdf:description></foo>";
const string rdf_result = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><bare xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"><rdf:description>ipsum lorum</rdf:description></bare>";

BOOST_AUTO_TEST_CASE(test_namespaces)
{
	TIODocument doc(bare);
	Namespaces ns(doc->getNamespaces());
	ostringstream xml;

	// correct usage tests
	ns.put("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
	xml << doc;
	BOOST_CHECK_EQUAL(rdf_bare, xml.str());

	Location l(xml.str(), "/bare");
	Location desc(rdf_desc, "//rdf:description");
	desc = "ipsum lorum";
	l.add(desc);
	xml.str("");
	xml << l.getDocument();
	BOOST_CHECK_EQUAL(rdf_result, xml.str());

	// adding second theme should be no-op
	ns.put("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");

	// bad usage tests
	BOOST_CHECK_THROW(ns.put("rdf", "yadda"), XMLTIO_Exception);
}

BOOST_AUTO_TEST_CASE(test_namespaces_merge)
{
	TIODocument doc(bare);
	Namespaces ns(doc->getNamespaces());
	TIODocument rdf_doc(rdf_bare);
	ns.merge(rdf_doc->getNamespaces());

	BOOST_CHECK_EQUAL(rdf_bare, dom2string(doc));
}

BOOST_AUTO_TEST_CASE(test_add_w_namespace)
{
	Location lb(bare, "/bare");
	Location lr(bare_plus_foo_ns, "//xcf:foo");

	lb.add(lr);
	BOOST_CHECK_EQUAL(bare_plus_foo_ns, lb.getDocumentText());
}

string create_element(const std::string& uri,
                      const std::string& prefix,
                      const std::string& qname)
{
	TIODocument doc(bare);
	Location lb(doc, "/bare");
	ostringstream xml;

	Location l(doc.createElement(uri, prefix, qname));
	lb.add(l);

	xml << doc;
	return xml.str();
}

BOOST_AUTO_TEST_CASE(test_create_attribute_ns)
{
	TIODocument doc(bare);
	Location lb(doc, "/bare");

	Location attr(doc.createAttribute("http://purl.org/ActiveMemory",
	                                  "memory",
	                                  "memory:about"));
	attr = "about something";
	lb.add(attr);

	ostringstream xml;
	xml << doc;

	BOOST_CHECK_EQUAL(bare_memory_ns, xml.str());
}

BOOST_AUTO_TEST_CASE(test_remove_attribute)
{
	Location root(bare_memory_ns, "/*");
	Location about(root, "@memory:about");
	root.remove(about);

	BOOST_CHECK_EQUAL(bare_memory_ns_no_about,
	                  root.getDocumentText());
}

BOOST_AUTO_TEST_CASE(test_create_normal_element)
{
	string xml(create_element("", "", "foo"));
	BOOST_CHECK_EQUAL(bare_plus_foo, xml);
}

BOOST_AUTO_TEST_CASE(test_create_element_ns)
{
	string xml(create_element("http://purl.org/XCF", "xcf", "xcf:foo"));

	BOOST_CHECK_EQUAL(bare_plus_foo_ns, xml);
}

BOOST_AUTO_TEST_CASE(test_is_equal)
{
	Location l1(xmldata), l2(xmldata);
	BOOST_CHECK(l1.isEqual(l2));
}

BOOST_AUTO_TEST_CASE(test_replace_content)
{
	Location args(xmldata, "/xcf/arguments");
	Location item(xml_attr, "/item");

	args = LocationContent(item);
	BOOST_CHECK_EQUAL(itemdata, args.getDocumentText());
}

BOOST_AUTO_TEST_CASE(test_replace_root)
{
	Location xd(xmldata, "/*");
	Location lb(bare, "/*");

	xd = LocationContent(lb);
	BOOST_CHECK_EQUAL(bare, xd.getDocumentText());
}

BOOST_AUTO_TEST_CASE(test_replace_root_by_location)
{
	Location xd(xmldata, "/*");
	Location lb(bare_plus_foo, "/bare/foo");

	xd = LocationContent(lb);
	BOOST_CHECK_EQUAL(just_foo, xd.getDocumentText());
}

BOOST_AUTO_TEST_CASE(test_document_clone)
{
	Location xd(xmldata, "/");
	TIODocument newdoc=xd.getDocument().clone();
	Location newloc(newdoc,"/");
	// change the original location, should leave the clone uneffected
	xd[XPath("/xcf/arguments/@arg")]="k";
	BOOST_CHECK_EQUAL(xmldata, newloc.getDocumentText());
}

BOOST_AUTO_TEST_CASE(test_extract_long)
{
	Location xd("<A value=\"2395388042\"/>","/A/@value");
	long long int l=(long long) extract<long long>(xd);
	BOOST_CHECK_EQUAL(l,(long long)2395388042l);
}


BOOST_AUTO_TEST_CASE(test_check_version)
{
	BOOST_CHECK(check_xmltio_version(XMLTIO_MAJOR, XMLTIO_MINOR,
	                                 XMLTIO_ABI_VERSION));
	BOOST_CHECK(!check_xmltio_version(XMLTIO_MAJOR -1,
	                                  XMLTIO_MINOR -1,
	                                  XMLTIO_ABI_VERSION -1));
}

BOOST_AUTO_TEST_CASE(test_loc2doc)
{
	Location l("<foo><bar attr=\"bla\"><child/></bar></foo>", "/foo/bar");
	TIODocument d(l);
	ostringstream oss;
	oss << d;
	BOOST_CHECK_EQUAL("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><bar attr=\"bla\"><child/></bar>", oss.str());
}

BOOST_AUTO_TEST_CASE(test_createAPI)
{
	std::string xml("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?><foo><bar/><bar/><bar><foo arg1=\"1\" arg2=\"2\"/></bar></foo>");
	Location root("<foo/>", "/foo");
	root.createElement("bar");
	root.createElement("bar");
	Location l = root.createElement("bar");
	l = l.createElement("foo");
	l["arg1"] = 1;
	l["arg2"] = 2;
	BOOST_CHECK_EQUAL(root.getDocumentText(),xml);
}

BOOST_AUTO_TEST_CASE(test_exist)
{
	Location xml("<list><foo arg='foo'/><foo arg='bar'/><foo arg='foo'/><foo att='bar'/></list>", "/");
	XPathIterator i = XPath("//foo[not(@arg) or @arg!='foo']").evaluate(xml);
	BOOST_CHECK(i);
	++i; BOOST_CHECK(i);
	++i; BOOST_CHECK(!i);
}
